; Kmmel Mandelbrot Benchmark v0.53 MT by Michael Kbel and F0dder 10/07/2006
format PE GUI 4.0
entry WINMAIN

include '%fasminc%\win32a.inc'
include '%fasminc%\macro\if.inc'
include '%fasminc%\ddraw.inc'

ALGO = 1		; This declares which algoritm you wanna compile
				; 0 = FPU
				; 1 = SSE2
				; 2 = SSE3
				; 3 = SSE3 + Vodnaya

LINE_INTERLEAVE = 1 ; this says how many lines one thread should calculate at a time

REPEAT_NUMBER = 5	; this specifies how much times the benchmark is repeated due to short execution time on fast CPU's
REPEAT_SHIFT = 2    ; this is the value the iterations and time are shifted right due to the exceeding of the max 32 bt number,
					;...so the REPEAT_NUMBER has to be more or less about equal to 2^REPEAT_SHIFT
					;...sorry for this hack, but it's kind of convenient ;-)

SCREENWIDTH		=	800
SCREENHEIGHT	=	600
WINDOWSIZE		=	600		; Set Size of Calcualtion Area (always quadratic)
ITERATIONS		=	2400	; Iterations for fractal, is fit to the colour palette of 240 entries

macro FrameArray [rzl,rzh,izl]
{
forward
	dq rzl
	dq rzh
	dq izl
}

macro ColourArray [t,r,g,b]
{
forward
	db r
	db b
	db g
	db t
}

section '.code' code readable executable

proc WINMAIN hinst, hprevinst, lpcmdline, ncmdshow
	; Define a WNDCLASS structure for a Windows window.
	invoke	GetModuleHandle, 0
	mov		[hinstance], eax
	invoke	LoadIcon, 0, IDI_APPLICATION
	mov		[wc.hIcon], eax
	invoke	LoadCursor, 0, IDC_ARROW
	mov		[wc.hCursor], eax
	mov		[wc.style], CS_DBLCLKS or CS_OWNDC or CS_HREDRAW or CS_VREDRAW
	mov		[wc.lpfnWndProc], WindowProc
	mov		[wc.cbClsExtra], 0
	mov		[wc.cbWndExtra], 0
	mov		eax, [hinstance]
	mov		[wc.hInstance], eax
	invoke	GetStockObject, BLACK_BRUSH
	mov		[wc.hbrBackground], eax
	mov		[wc.lpszMenuName], 0
	mov		[wc.lpszClassName], _class

	; Register our window class.
	invoke	RegisterClass,wc

	; Create our window in the system, and make it viewable.
	invoke	CreateWindowEx, 0, _class, _title,\
				WS_POPUP+WS_VISIBLE, 100, 30, WINDOWSIZE, WINDOWSIZE, 0, 0,\
				[hinstance], 0
	mov		[mainhwnd], eax
	
	; zero out directdraw object pointers
	mov		[DDraw],0
	mov		[lpddsprimary],0
	mov		[lpddsback],0
	mov		[lpddpal],0

	invoke	DirectDrawCreate, 0, DDraw, 0

	cominvk DDraw, SetCooperativeLevel, [mainhwnd], DDSCL_FULLSCREEN or DDSCL_EXCLUSIVE

	; Change the display mode
	cominvk DDraw, SetDisplayMode, SCREENWIDTH, SCREENHEIGHT, 32, 0, 0

	mov		[ddsd.dwSize], sizeof.DDSURFACEDESC
	mov		[ddsd.dwFlags], DDSD_CAPS
	mov		[ddsd.ddsCaps.dwCaps], DDSCAPS_PRIMARYSURFACE

	; Create directdraw primary surface
	cominvk DDraw, CreateSurface, ddsd, lpddsprimary, 0

	mov		[ddsd.dwSize], sizeof.DDSURFACEDESC

;--------------------------------------------------------------------

	; Init Colour Table (fits to 2400 ITERATIONS for speeding up the lookup table)
	mov		esi, Colour_Table_init
	mov		edi, Colour_Table
	mov		eax, 10 							; 2400 ITERATIONS/240
colour_table_copy_loop1:
	xor		ebx, ebx
colour_table_copy_loop2:
	mov		ecx, [esi+ebx]
	mov		[edi+ebx], ecx
	add		ebx, 4
	cmp		ebx, (240*4)
	jne		colour_table_copy_loop2

	add		edi, (240*4)
	dec		eax
	jnz		colour_table_copy_loop1

	mov		dword[edi], 00000000h				; last colour (Maximum ITERATIONS is black !)


	; Disable Mouse and Keyboard Handle to Window
	invoke	EnableWindow, [mainhwnd], 0
	
	; <f0dder> moved the locking to "as late as possible"
	; lock surface for drawing directly to frame buffer
	cominvk lpddsprimary, LockSurface, 0, ddsd, DDLOCK_SURFACEMEMORYPTR or DDLOCK_WAIT, 0

	; boost process priority class for even more correct results
	invoke	GetCurrentProcess
	invoke	SetPriorityClass, eax, REALTIME_PRIORITY_CLASS

	; Get the timer of the System
	invoke	GetTickCount
	mov		[timer], eax

	;
	; <f0dder> - threading modifications begin
	;
	mov		[repeat_counter], REPEAT_NUMBER
	
	repeat_counter_loop:
	
		mov		[frame_counter], MAX_FRAMES
 
		frame_counter_loop:

		;Thread Core 1
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*0, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*0, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*0, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*0, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 0], eax

		;Thread Core 2
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*1, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*1, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*1, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*1, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 4], eax

		;Thread Core 3
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*2, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*2, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*2, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*2, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 8], eax

		;Thread Core 4
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*3, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*3, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*3, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*3, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 12], eax

		;Thread Core 5
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*4, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*4, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*4, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*4, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 16], eax
		;Thread Core 6
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*5, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*5, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*5, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*5, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 20], eax
		;Thread Core 7
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*6, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*6, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*6, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*6, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 24], eax
		;Thread Core 8
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*7, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*7, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*7, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*7, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 28], eax
		;Thread Core 9
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*8, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*8, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*8, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*8, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 32], eax
		;Thread Core 10
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*9, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*9, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*9, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*9, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 36], eax
		;Thread Core 11
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*10, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*10, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*10, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*10, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 40], eax
		;Thread Core 12
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*11, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*11, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*11, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*11, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 44], eax
		;Thread Core 13
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*12, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*12, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*12, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*12, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 48], eax
		;Thread Core 14
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*13, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*13, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*13, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*13, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 52], eax
		;Thread Core 15
		
		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*14, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*14, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*14, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*14, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 56], eax
		
		;Thread Core 16

		if ALGO=0
			invoke	CreateThread,NULL, 0, thread_draw_fpu, LINE_INTERLEAVE*15, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=1
			invoke	CreateThread,NULL, 0, thread_draw_sse2, LINE_INTERLEAVE*15, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=2
			invoke	CreateThread,NULL, 0, thread_draw_sse3, LINE_INTERLEAVE*15, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		if ALGO=3
			invoke	CreateThread,NULL, 0, thread_draw_sse3_vodnaya, LINE_INTERLEAVE*15, \
					REALTIME_PRIORITY_CLASS or CREATE_SUSPENDED, tId
		end if
		mov		[threadhandles + 60], eax
		
		mov		[current_line], LINE_INTERLEAVE*15				; start drawing here next
				
		invoke	SetThreadAffinityMask, [threadhandles +  0],0000000000000001b 
		invoke	SetThreadAffinityMask, [threadhandles +  4],0000000000000010b 
		invoke	SetThreadAffinityMask, [threadhandles +  8],0000000000000100b 
		invoke	SetThreadAffinityMask, [threadhandles + 12],0000000000001000b 
		invoke	SetThreadAffinityMask, [threadhandles + 16],0000000000010000b 
		invoke	SetThreadAffinityMask, [threadhandles + 20],0000000000100000b 
		invoke	SetThreadAffinityMask, [threadhandles + 24],0000000001000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 28],0000000010000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 32],0000000100000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 36],0000001000000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 40],0000010000000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 44],0000100000000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 48],0001000000000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 52],0010000000000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 56],0100000000000000b 
		invoke	SetThreadAffinityMask, [threadhandles + 60],1000000000000000b 

		invoke	ResumeThread, [threadhandles +  0]
		invoke	ResumeThread, [threadhandles +  4]
		invoke	ResumeThread, [threadhandles +  8]
		invoke	ResumeThread, [threadhandles + 12]
		invoke	ResumeThread, [threadhandles + 16]
		invoke	ResumeThread, [threadhandles + 20]
		invoke	ResumeThread, [threadhandles + 24]
		invoke	ResumeThread, [threadhandles + 28]
		invoke	ResumeThread, [threadhandles + 32]
		invoke	ResumeThread, [threadhandles + 36]
		invoke	ResumeThread, [threadhandles + 40]
		invoke	ResumeThread, [threadhandles + 44]
		invoke	ResumeThread, [threadhandles + 48]
		invoke	ResumeThread, [threadhandles + 52]
		invoke	ResumeThread, [threadhandles + 56]
		invoke	ResumeThread, [threadhandles + 60]

		; Wait until both threads are done rendering
		invoke	WaitForMultipleObjectsEx,16, threadhandles, 1, -1, 0
		invoke	CloseHandle, [threadhandles +  0]
		invoke	CloseHandle, [threadhandles +  4]
		invoke	CloseHandle, [threadhandles +  8]
		invoke	CloseHandle, [threadhandles + 12]
		invoke	CloseHandle, [threadhandles + 16]
		invoke	CloseHandle, [threadhandles + 20]
		invoke	CloseHandle, [threadhandles + 24]
		invoke	CloseHandle, [threadhandles + 28]
		invoke	CloseHandle, [threadhandles + 32]
		invoke	CloseHandle, [threadhandles + 36]
		invoke	CloseHandle, [threadhandles + 40]
		invoke	CloseHandle, [threadhandles + 44]
		invoke	CloseHandle, [threadhandles + 48]
		invoke	CloseHandle, [threadhandles + 52]
		invoke	CloseHandle, [threadhandles + 56]
		invoke	CloseHandle, [threadhandles + 60]

		dec		[frame_counter]
		jns		frame_counter_loop
	
		mov		eax,[global_iter_count]       
		shr		eax,REPEAT_SHIFT		;That has to be done, because counter_iter exceeds the 32bit integer limit
		mov		[global_iter_count],0
		add		[counter_iter_ges],eax
	
	dec		[repeat_counter]
	jnz		repeat_counter_loop

	;
	; <f0dder> - threading modifications ned
	;

	invoke	GetTickCount
	sub		eax,[timer]
	shr		eax,REPEAT_SHIFT	;That has to be done, because counter_iter exceeds the 32bit integer limit and so the calculation has to be adapted
	mov		[timer],eax
	
	fild	[counter_iter_ges]
	fidiv	[timer]
	fistp	dword[timer]	;"Calculate Iteraions/Second
	
	; Enable Mouse and Keyboard Handle to Window
	invoke	EnableWindow,[mainhwnd],1
	
	; ALWAYS! unlock surface when finished drawing
	cominvk lpddsprimary,UnlockSurface,NULL

	;------------------------------------------------------------------------
	; All the fun drawing stuff is done now, time to show results
	;------------------------------------------------------------------------

	; Preparation for number to string conversion for the timer result
	mov [address_text], text_result+38 ;You can add constants together here @Madis@

	; Conversion of number to string conversion for the timer result
NumberToString_loop:				;  Totally rewritten @Madis@
	mov		eax, [timer]			; Our timer is in eax and it remains there
	test	eax, eax
	jne		.overzeromilliseconds
	mov		esi, [address_text]
	mov		dword[esi], '0  '		; Could be used to put additional text like '0 s' or so...
	jmp		.underonesecond
.overzeromilliseconds:
	mov		edi, buffer				; I reserved another buffer where we put our intermediate result in
	xor		ecx, ecx
	mov		esi, 10 				; Esi hold the 10 we are going to divide with
    .anotherchar:
	test	eax, eax				; Lets test eax for zero
	je		.end					; We are done if it holds true
	xor		edx, edx
	div		esi						; divide EDX:EAX with ESI putting the result in EAX and modulo in EDX
	mov		[edi+ecx], dl			; dl is in range 0..9
	inc		ecx						; increase the pointer by one
	jmp		.anotherchar
.end:
	mov		esi, [address_text]		; We don't need esi so we are reusing it for text pointer
	mov		eax, [edi]				; Take the first 4 bytes from the buffer
	add		eax, "0000"				; make them number strings
	bswap	eax						; swap them because they are in reverse order
									; 3145 will be 5413 after this instruction
	mov		al, '.' 				; We need only three numbers, the 4th is replaced with dot
	sub		ecx, 3
	jnc		@f
	mov		ecx, 1					; If ecx gets too small then correct it
@@:
	mov		dword[esi+ecx], eax		; We start filling from the end
	mov		word[esi+ecx+4], '  '	; Lets add seconds too: example 'Z.123  ' Z means empty
	add		edi, 3
	mov		byte[esi+ecx-1], "0"	; Add a nice zero so we get: '0.123  '                 <=
	test	ecx,ecx 				; If the result is under one second then skip to end
	je		.underonesecond
.putonebyte:
	mov		al, [edi]				; Here we take bytes from [edi]...
	add		al, "0" 				; ...add 30h to them and...
	mov		[esi+ecx-1], al 		; ...finally write it back to where we were 6 lines ago ^
	inc		edi						; This way '15.123  ' will replace '0.123  '
	loop	.putonebyte
.underonesecond:					; End of rewriting @Madis@

end_of_progi:
	;free the directdraw surface object
	.if dword [lpddsprimary], ne, 0
		cominvk lpddsprimary, Release
	.endif
	
	;free the directdraw object
	.if dword [DDraw], ne, 0
		cominvk DDraw, Release
	.endif
	
	invoke	CloseWindow, [mainhwnd]
	invoke	MessageBox, HWND_DESKTOP, text_result, caption, [flags]
	invoke	ExitProcess, [msg.wParam]
	
	ret 0
endp

;--------------------------------------------------------------------
; The procedure called when Windows has a message for your program
;--------------------------------------------------------------------
proc WindowProc  hwnd, wmsg, wparam, lparam
	mov		eax, [wmsg]

	.if eax, e, WM_CLOSE
		invoke	DestroyWindow,[hwnd]
		ret	0
	.elseif eax, e, WM_DESTROY
		invoke	PostQuitMessage, 0
		ret	0
	.elseif eax, e, WM_PAINT
		invoke	BeginPaint, [hwnd], ps
		invoke	EndPaint, [hwnd], ps
		ret		0
	.endif

	invoke	DefWindowProc, [hwnd], [wmsg], [wparam], [lparam]
	ret
endp

;--------------------------------------------------------------------
; Now things get exciting: this is the fractal rendering code
;--------------------------------------------------------------------
proc thread_draw_fpu uses ebx esi, plot_y:DWORD
	local	rz_low:QWORD,			rz_high:QWORD,\	
			iz_low:QWORD,			iz:QWORD,\
			rz:QWORD, 				dz:QWORD,\
			plot_x:DWORD,			plot_limit:DWORD,\ 
			local_iter_count:DWORD, dummy:DWORD
	;this order of the variables seems to be the best regarding alignment of ebp/esp and also the variables themselves,
	;the dummy is needed for the correct alignment of the data overall , like memory alloc for a memory amound aligned to 16
	
	mov		eax, [frame_counter]				; "copy the desired data for the frame
	imul	eax, 24 							; "address =  3 * 8 * framecounter

	fld		qword [frameconfig + 0	+ eax]
	fld		qword [frameconfig + 8	+ eax]
	fld		qword [frameconfig + 16 + eax]
	fstp	[iz_low]
	fstp	[rz_high]
	fstp	[rz_low]

	fld		[rz_high]							; dz = abs(rz_high - rz_low)/screen_size
	fsub	[rz_low]
	fabs
	fidiv	[window_size]
	fstp	[dz]
	
	mov     [local_iter_count],0
	mov		edx, Colour_Table					; pointer to colour table
.main_loop:
	mov		ebx, [ddsd.lpSurface]				; get address to surface
	add		ebx, (SCREENWIDTH-WINDOWSIZE)/2*4	; place it in the middle of the screen

	; Calculate scanline offset to draw to
	mov		esi, [ddsd.lPitch]
	mov		eax, [plot_y]
	imul	eax, esi
	add		ebx, eax
	sub		esi, WINDOWSIZE*4

	; calculate plotlimit
	mov		eax, [plot_y]
	add		eax, LINE_INTERLEAVE
	mov		[plot_limit], eax

.y_loop:
	fld		[dz]								; iz = dz * plot_y + iz_low
	fimul	[plot_y]
	fadd	[iz_low]
	fstp	[iz]
	
	mov		[plot_x], 0
.x_loop:
	fcompp		; clear 3*2 registers (according to optimisation guide fcompp is best) to clear FPU registers
	fcompp		; could be avoided somehow I think by better organisation of the registers/rz/iz change
	fcompp

	fld		[limit]
	
	fld		[dz]
	mov		eax, ITERATIONS
	fimul	[plot_x]
	fadd	[rz_low]						; rz = dz * plot_x + rz_low   -> better precision than adding...
	
	fld		[iz]							; could be avoided somehow...just the iteration_loop is more critical...
	fld		st1
	fld		st1
.iteration_loop:
	fld		st1
	fmul	st0,st2
	fld		st1
	fmul	st0,st2
	fxch	st2
	fmulp	st3,st0
	fld		st0
	fadd	st0,st2
	fxch	st3
	fadd	st0,st0
	fxch	st2
	fsubp	st1,st0
	fxch	st2
	fcomip	st5
	fadd	st0,st2
	fxch	st1
	fadd	st0,st3
	fxch	st1
	ja		.end_of_iteration

	sub		eax,1
	jnz		.iteration_loop
.end_of_iteration:
	
	; accumulate number of iterations
	mov		ecx, ITERATIONS
	sub		ecx,eax
	add		[local_iter_count], ecx				

	; get color word
	shl		ecx, 2
	mov		eax, [edx+ecx]
	mov		[ebx], eax
	
	add		[plot_x],1
	add		ebx, 4
	
	cmp		[plot_x], WINDOWSIZE
	jne		.x_loop
	
	add		[plot_y],1
	add		ebx, esi

	mov		eax, [plot_limit]
	cmp		[plot_y], eax
	jne		.y_loop

	; update the global current_line counter atomically
	mov		eax, LINE_INTERLEAVE
	lock	xadd [current_line], eax
	add		eax, LINE_INTERLEAVE
	mov		[plot_y], eax

	; keep drawing if there's more to do
	cmp		eax, WINDOWSIZE
	jl		.main_loop	

	mov		eax,[local_iter_count]
	lock	add [global_iter_count],eax

	ret
endp

;--------------------------------------------------------------------
; And now in SSE2 version
;--------------------------------------------------------------------
proc thread_draw_sse2 uses ebx esi, plot_y:DWORD
	local	rz_low:QWORD,			rz_high:QWORD,\
			iz_low:QWORD,			dz:QWORD,\
			iz_temp_fpu:QWORD, 		rz_temp_fpu:QWORD,\
			plot_x:DWORD, 			plot_limit:DWORD,\
			local_iter_count:DWORD,	dummy:DWORD
	;this order of the variables seems to be the best regarding alignment of ebp/esp and also the variables themselves,
	;the dummy is needed for the correct alignment of the data overall , like memory alloc for a memory amound aligned to 16

	mov		eax, [frame_counter]				; "copy the desired data for the frame
	imul	eax, 24 							; "address =  3 * 8 * framecounter

	fld		qword [frameconfig + 0	+ eax]
	fld		qword [frameconfig + 8	+ eax]
	fld		qword [frameconfig + 16 + eax]
	fstp	[iz_low]
	fstp	[rz_high]
	fstp	[rz_low]

	fld		[rz_high]							; dz = abs(rz_high - rz_low)/screen_size
	fsub	[rz_low]
	fabs
	fidiv	[window_size]
	fstp	[dz]
	
	movapd	xmm7, dqword [radiant]				; get constant mask for radiant (4.0 4.0)

	mov     [local_iter_count],0
	mov		edx, Colour_Table					; pointer to colour table

.main_loop:
	mov		ebx, [ddsd.lpSurface]				; get address to surface
	add		ebx, (SCREENWIDTH-WINDOWSIZE)/2*4	; center horizontally

	; Calculate scanline offset to draw to
	mov		esi, [ddsd.lPitch]
	mov		eax, [plot_y]
	imul	eax, esi
	add		ebx, eax
	sub		esi, WINDOWSIZE*4					; center horizontally

	; calculate plotlimit
	mov		eax, [plot_y]
	add		eax, LINE_INTERLEAVE
	mov		[plot_limit], eax

.y_loop:
	; iz = dz / screen_size * plot_y + iz_low -> better precision than adding...
	fld		[dz]
	fimul	[plot_y]
	fadd	[iz_low]
	fstp	[iz_temp_fpu]
	
	movsd	xmm5, [iz_temp_fpu]
	shufpd	xmm5, xmm5, 0						; xmm5:    iz   |  iz
	
	mov		[plot_x], 0
.x_loop:
	; rz = dz / screen_size * plot_x + rz_low -> better precision than adding...
	fld		[dz]
	fimul	[plot_x]
	fadd	[rz_low]
	fstp	[rz_temp_fpu]
	
	xorpd	xmm2, xmm2							; xmm2:    0    |   0       (zeros xmm6)
	movlpd	xmm4, [rz_temp_fpu]					; xmm4:    rz   |   ?       (moves rz to low QW)
	movhpd	xmm2, [dz]							; xmm2:    0    |   dz      (moves dz in high QW, low QW is unchanged)
	shufpd	xmm4, xmm4, 0						; xmm4:    rz   |   rz      (because of '0' the low QW is moved to high QW)
	addpd	xmm4, xmm2							; xmm4:    rz   |   rz + dz (sub to fit to the plot_x/y values...)
	
	movapd	xmm0, xmm4							; xmm0:    rz   |   rz + dz (transfer from one to other reg)
	xorpd	xmm6, xmm6							; xmm6:    0    |   0       (zeros xmm6)
	movapd	xmm1, xmm5							; xmm1:    iz   |   iz
	
	mov		ecx, ITERATIONS

.iteration_loop:
	movapd	xmm2, xmm0							; xmm2:    rz           |   rz + dz
	mulpd	xmm0, xmm0							; xmm0:    rz^2         |   (rz + dz)^2
	movapd	xmm3, xmm1							; xmm3:    iz           |   iz
	addpd	xmm1, xmm1							; xmm1:    iz+iz        |   iz+iz
	mulpd	xmm1, xmm2							; xmm1:    2*iz*rz      |   2*iz*(rz+dz)
	movapd	xmm2, xmm0							; xmm2:    rz^2         |   rz^2
	mulpd	xmm3, xmm3							; xmm3:    iz^2         |   iz^2
	
	addpd	xmm1, xmm5							; xmm1:    2*iz*rz+iz0  |   2*iz*(rz+dz)+iz0
	subpd	xmm0, xmm3							; xmm0:    rz^2-iz^2    |   (rz-dz)^2-iz^2
	addpd	xmm2, xmm3							; xmm2:    rz^2+iz^2    |   (rz-dz)^2+iz^2
	
	cmplepd xmm2, xmm7							; xmm2 <=  4.0 | 4.0 ? True -> QW = FFFFFFFFFFFFFFFFh else 0000000000000000h
	addpd	xmm0, xmm4							; xmm0:    rz^2-iz^2+rz0|   (rz-+dz)^2-iz^2+rz0
	movmskpd	eax, xmm2						; get the sign bits of the two QW in xmm2 in eax -> so either 00,01,10,11

	test	eax, eax
	jz		.end_of_iteration

	andpd	xmm2, xmm7							; get either 4.0 or 0.0 for each iteration number
	addpd	xmm6, xmm2							; add        4.0 or 0.0 to the iteration counter
	sub		ecx, 1
	jnz		.iteration_loop
	
.end_of_iteration:
	cvttsd2si	ecx, xmm6						; convert low QW of xmm6 into signed integer with truncation
	mov		eax, [edx+ecx]						; get colour word
	shr		ecx, 2								; div by 4 to get right number of iterations
	mov		[ebx], eax							; Plot pixel
	add		[local_iter_count],ecx					; save number of iterations and accumulate for calculation
	
	shufpd	xmm6, xmm6, 1						; move high QW to low QW in xmm6
	cvttsd2si	ecx, xmm6						; convert low QW of xmm6 into signed integer with truncation
	mov		eax, [edx+ecx]						; get colour word
	shr		ecx, 2								; div by 4 to get right number of iterations
	mov		[ebx+4], eax						; Plot pixel
	add		[local_iter_count], ecx					; save number of iterations and accumulate for calculation
	
	add		[plot_x], 2
	add		ebx, 8
	
	cmp		[plot_x], WINDOWSIZE
	jne		.x_loop
	
	add		[plot_y],1
	add		ebx,esi

	mov		eax, [plot_limit]
	cmp		[plot_y], eax
	jne		.y_loop

	; update the global current_line counter atomically
	mov		eax, LINE_INTERLEAVE
	lock	xadd [current_line], eax
	add		eax, LINE_INTERLEAVE
	mov		[plot_y], eax

	; keep drawing if there's more to do
	cmp		eax, WINDOWSIZE
	jl		.main_loop	
	
	mov		eax,[local_iter_count]
	lock	add [global_iter_count],eax
	ret
endp

;--------------------------------------------------------------------
; And now in SSE3 version
;--------------------------------------------------------------------
proc thread_draw_sse3 uses ebx esi, plot_y:DWORD
	local	rz_low:QWORD,			rz_high:QWORD,\
			iz_low:QWORD,			dz:QWORD,\
			iz_temp_fpu:QWORD, 		rz_temp_fpu:QWORD,\
			plot_x:DWORD, 			plot_limit:DWORD,\
			local_iter_count:DWORD,	dummy:DWORD
	;this order of the variables seems to be the best regarding alignment of ebp/esp and also the variables themselves,
	;the dummy is needed for the correct alignment of the data overall , like memory alloc for a memory amound aligned to 16

	mov		eax, [frame_counter]				; "copy the desired data for the frame
	imul	eax, 24 							; "address =  3 * 8 * framecounter

	fld		qword [frameconfig + 0	+ eax]
	fld		qword [frameconfig + 8	+ eax]
	fld		qword [frameconfig + 16 + eax]
	fstp	[iz_low]
	fstp	[rz_high]
	fstp	[rz_low]

	fld		[rz_high]							; dz = abs(rz_high - rz_low)/screen_size
	fsub	[rz_low]
	fabs
	fidiv	[window_size]
	fstp	[dz]
	
	movapd	xmm7, dqword [radiant]				; get constant mask for radiant (4.0 4.0)
	mov     [local_iter_count],0
	mov		edx, Colour_Table					; pointer to colour table

.main_loop:
	mov		ebx, [ddsd.lpSurface]				; get address to surface
	add		ebx, (SCREENWIDTH-WINDOWSIZE)/2*4	; center horizontally

	; Calculate scanline offset to draw to
	mov		esi, [ddsd.lPitch]
	mov		eax, [plot_y]
	imul	eax, esi
	add		ebx, eax
	sub		esi, WINDOWSIZE*4					; center horizontally

	; calculate plotlimit
	mov		eax, [plot_y]
	add		eax, LINE_INTERLEAVE
	mov		[plot_limit], eax

.y_loop:
	; iz = dz / screen_size * plot_y + iz_low -> better precision than adding...
	fld		[dz]
	fimul	[plot_y]
	fadd	[iz_low]
	fstp	[iz_temp_fpu]
	
	mov		[plot_x], 0
	movddup	xmm5, [iz_temp_fpu]				; xmm5:    iz   |  iz    (SSE3) 
	
.x_loop:
	; rz = dz / screen_size * plot_x + rz_low -> better precision than adding...
	fld		[dz]
	fimul	[plot_x]
	fadd	[rz_low]
	fstp	[rz_temp_fpu]
	
	xorpd	xmm2, xmm2							; xmm2:    0    |   0       (zeros xmm6)
	movddup	xmm4, [rz_temp_fpu]					; xmm4:    rz   |   rz     (SSE3) 
	movhpd	xmm2, [dz]							; xmm2:    0    |   dz      (moves dz in high QW, low QW is unchanged)
	movapd	xmm1, xmm5							; xmm1:    iz   |   iz
	addpd	xmm4, xmm2							; xmm4:    rz   |   rz + dz (sub to fit to the plot_x/y values...)
	xorpd	xmm6, xmm6							; xmm6:    0    |   0       (zeros xmm6)
	movapd	xmm0, xmm4							; xmm0:    rz   |   rz + dz (transfer from one to other reg)
	
	mov		ecx, ITERATIONS

.iteration_loop:
	movapd	xmm2, xmm0							; xmm2:    rz           |   rz + dz
	mulpd	xmm0, xmm0							; xmm0:    rz^2         |   (rz + dz)^2
	movapd	xmm3, xmm1							; xmm3:    iz           |   iz
	addpd	xmm1, xmm1							; xmm1:    iz+iz        |   iz+iz
	mulpd	xmm1, xmm2							; xmm1:    2*iz*rz      |   2*iz*(rz+dz)
	movapd	xmm2, xmm0							; xmm2:    rz^2         |   rz^2
	mulpd	xmm3, xmm3							; xmm3:    iz^2         |   iz^2
	
	addpd	xmm1, xmm5							; xmm1:    2*iz*rz+iz0  |   2*iz*(rz+dz)+iz0
	subpd	xmm0, xmm3							; xmm0:    rz^2-iz^2    |   (rz-dz)^2-iz^2
	addpd	xmm2, xmm3							; xmm2:    rz^2+iz^2    |   (rz-dz)^2+iz^2
	
	cmplepd xmm2, xmm7							; xmm2 <=  4.0 | 4.0 ? True -> QW = FFFFFFFFFFFFFFFFh else 0000000000000000h
	addpd	xmm0, xmm4							; xmm0:    rz^2-iz^2+rz0|   (rz-+dz)^2-iz^2+rz0
	movmskpd	eax, xmm2						; get the sign bits of the two QW in xmm2 in eax -> so either 00,01,10,11

	test	eax, eax
	jz		.end_of_iteration

	andpd	xmm2, xmm7							; get either 4.0 or 0.0 for each iteration number
	addpd	xmm6, xmm2							; add        4.0 or 0.0 to the iteration counter
	sub		ecx, 1
	jnz		.iteration_loop
	
.end_of_iteration:
	cvttsd2si	ecx, xmm6						; convert low QW of xmm6 into signed integer with truncation
	mov		eax, [edx+ecx]						; get colour word
	shr		ecx, 2								; div by 4 to get right number of iterations
	mov		[ebx], eax							; Plot pixel
	add		[local_iter_count],ecx					; save number of iterations and accumulate for calculation
	
	shufpd	xmm6, xmm6, 1						; move high QW to low QW in xmm6
	cvttsd2si	ecx, xmm6						; convert low QW of xmm6 into signed integer with truncation
	mov		eax, [edx+ecx]						; get colour word
	shr		ecx, 2								; div by 4 to get right number of iterations
	mov		[ebx+4], eax						; Plot pixel
	add		[local_iter_count], ecx					; save number of iterations and accumulate for calculation
	
	add		[plot_x], 2
	add		ebx, 8
	
	cmp		[plot_x], WINDOWSIZE
	jne		.x_loop
	
	add		[plot_y],1
	add		ebx,esi

	mov		eax, [plot_limit]
	cmp		[plot_y], eax
	jne		.y_loop

	; update the global current_line counter atomically
	mov		eax, LINE_INTERLEAVE
	lock	xadd [current_line], eax
	add		eax, LINE_INTERLEAVE
	mov		[plot_y], eax

	; keep drawing if there's more to do
	cmp		eax, WINDOWSIZE
	jl		.main_loop	
	
	mov		eax,[local_iter_count]
	lock	add [global_iter_count],eax
	ret
endp

;--------------------------------------------------------------------
; And now in SSE3 Vodnaya version
;--------------------------------------------------------------------
proc thread_draw_sse3_vodnaya uses ebx esi, plot_y:DWORD
	local	rz_low:QWORD,			rz_high:QWORD,\
			iz_low:QWORD,			dz:QWORD,\
			iz_temp_fpu:QWORD, 		rz_temp_fpu:QWORD,\
			plot_x:DWORD, 			plot_limit:DWORD,\
			local_iter_count:DWORD,	dummy:DWORD
	;this order of the variables seems to be the best regarding alignment of ebp/esp and also the variables themselves,
	;the dummy is needed for the correct alignment of the data overall , like memory alloc for a memory amound aligned to 16

	mov		eax, [frame_counter]				; "copy the desired data for the frame
	imul	eax, 24 							; "address =  3 * 8 * framecounter

	fld		qword [frameconfig + 0	+ eax]
	fld		qword [frameconfig + 8	+ eax]
	fld		qword [frameconfig + 16 + eax]
	fstp	[iz_low]
	fstp	[rz_high]
	fstp	[rz_low]

	fld		[rz_high]							; dz = abs(rz_high - rz_low)/screen_size
	fsub	[rz_low]
	fabs
	fidiv	[window_size]
	fstp	[dz]
	
	movapd	xmm7, dqword [radiant]				; get constant mask for radiant (4.0 4.0)
	mov     [local_iter_count],0
	mov		edx, Colour_Table					; pointer to colour table

.main_loop:
	mov		ebx, [ddsd.lpSurface]				; get address to surface
	add		ebx, (SCREENWIDTH-WINDOWSIZE)/2*4	; center horizontally

	; Calculate scanline offset to draw to
	mov		esi, [ddsd.lPitch]
	mov		eax, [plot_y]
	imul	eax, esi
	add		ebx, eax
	sub		esi, WINDOWSIZE*4					; center horizontally

	; calculate plotlimit
	mov		eax, [plot_y]
	add		eax, LINE_INTERLEAVE
	mov		[plot_limit], eax

.y_loop:
	; iz = dz / screen_size * plot_y + iz_low -> better precision than adding...
	fld		[dz]
	fimul	[plot_y]
	fadd	[iz_low]
	fstp	[iz_temp_fpu]
	
	mov		[plot_x], 0
	movddup	xmm5, [iz_temp_fpu]				; xmm5:    iz   |  iz    (SSE3) 
	
.x_loop:
	; rz = dz / screen_size * plot_x + rz_low -> better precision than adding...
	fld		[dz]
	fimul	[plot_x]
	fadd	[rz_low]
	fstp	[rz_temp_fpu]
	
	xorpd	xmm2, xmm2							; xmm2:    0    |   0       (zeros xmm6)
	movddup	xmm4, [rz_temp_fpu]					; xmm4:    rz   |   rz     (SSE3) 
	movhpd	xmm2, [dz]							; xmm2:    0    |   dz      (moves dz in high QW, low QW is unchanged)
	movapd	xmm1, xmm5							; xmm1:    iz   |   iz
	addpd	xmm4, xmm2							; xmm4:    rz   |   rz + dz (sub to fit to the plot_x/y values...)
	xorpd	xmm6, xmm6							; xmm6:    0    |   0       (zeros xmm6)
	movapd	xmm0, xmm4							; xmm0:    rz   |   rz + dz (transfer from one to other reg)
	
	mov		ecx, ITERATIONS

.iteration_loop:
	movapd	xmm2, xmm1							; xmm2:    
	mulpd	xmm2, xmm2							; xmm2:  
	mulpd	xmm1, xmm0							; xmm1:  
	
	movapd	xmm3, xmm2							; xmm3:   
	mulpd	xmm0, xmm0							; xmm0:  
	addpd	xmm2, xmm0							; xmm2:    
	
	cmplepd xmm2, xmm7							; xmm2 <=  4.0 | 4.0 ? True -> QW = FFFFFFFFFFFFFFFFh else 0000000000000000h
	addpd	xmm1, xmm1
	subpd	xmm0, xmm3

	addpd	xmm1, xmm5							; xmm1:
	addpd	xmm0, xmm4							; xmm0:
	movmskpd	eax, xmm2						; get the sign bits of the two QW in xmm2 in eax -> so either 00,01,10,11
	test	eax, eax
	jz		.end_of_iteration

	andpd	xmm2, xmm7							; get either 4.0 or 0.0 for each iteration number
	addpd	xmm6, xmm2							; add        4.0 or 0.0 to the iteration counter
	sub		ecx, 1
	jnz		.iteration_loop
	
.end_of_iteration:
	cvttsd2si	ecx, xmm6						; convert low QW of xmm6 into signed integer with truncation
	mov		eax, [edx+ecx]						; get colour word
	shr		ecx, 2								; div by 4 to get right number of iterations
	mov		[ebx], eax							; Plot pixel
	add		[local_iter_count],ecx					; save number of iterations and accumulate for calculation
	
	shufpd	xmm6, xmm6, 1						; move high QW to low QW in xmm6
	cvttsd2si	ecx, xmm6						; convert low QW of xmm6 into signed integer with truncation
	mov		eax, [edx+ecx]						; get colour word
	shr		ecx, 2								; div by 4 to get right number of iterations
	mov		[ebx+4], eax						; Plot pixel
	add		[local_iter_count], ecx					; save number of iterations and accumulate for calculation
	
	add		[plot_x], 2
	add		ebx, 8
	
	cmp		[plot_x], WINDOWSIZE
	jne		.x_loop
	
	add		[plot_y],1
	add		ebx,esi

	mov		eax, [plot_limit]
	cmp		[plot_y], eax
	jne		.y_loop

	; update the global current_line counter atomically
	mov		eax, LINE_INTERLEAVE
	lock	xadd [current_line], eax
	add		eax, LINE_INTERLEAVE
	mov		[plot_y], eax

	; keep drawing if there's more to do
	cmp		eax, WINDOWSIZE
	jl		.main_loop	
	
	mov		eax,[local_iter_count]
	lock	add [global_iter_count],eax
	ret
endp

;------------------------------------------------------------------------------------------
section '.idata' import data readable writeable
;------------------------------------------------------------------------------------------
library kernel32,'KERNEL32.DLL',\
	user32,'USER32.DLL',\
	gdi32,'GDI32.DLL',\
	ddraw,'DDRAW.DLL'

include '%fasminc%/apia/kernel32.inc'
include '%fasminc%/apia/user32.inc'
include '%fasminc%/apia/gdi32.inc'

import ddraw,\
	DirectDrawCreate,'DirectDrawCreate'

;------------------------------------------------------------------------------------------
section '.data' data readable writeable
;------------------------------------------------------------------------------------------
  
msg MSG ;windows messages and information
wc  WNDCLASS ;describes how the window will look (its class)
ps  PAINTSTRUCT ;for general windows graphics

DDraw			DirectDraw						; DirectDraw calling methods (functions)
lpddsprimary	DirectDrawSurface				; DirectDraw primary surface
lpddsback		DirectDrawSurface				; DirectDraw backbuffer surface
lpddpal 		DirectDrawPalette				; Palette for 8-bit palletized surface

ddsd			DDSURFACEDESC					; Directdraw surface description
ddscaps 		DDSCAPS 						; Directdraw surface capabilities

align 16
radiant 			dq	4.0, 4.0		; radiant mask for SSE version
limit				dq	4.0				; dummy
global_iter_count	dd	0				; count of iterations
counter_iter_ges	dd	0				; total count of iterations of all frames

window_size		dd	WINDOWSIZE

timer				dd	0			; to store the timer at start
divisor 			dd	1000000000	; maximum divisor must be < 2^31...as far as I can know...
counter 			dd	10			; this makes 10 loops for conversion...
address_text		dd	0			; saves the pointer to text_result
flag_number_start	dd	0			; flag for the start of the number to get rid of leading zeros

text_result		db 'Speed [million iterations / second] : ',0
				rb 15	; reserve some 15 characters for the result string including end zero
						; Reserve by rb, rw, rd, etc and count to relieve typing
						; dup can also be used from other syntaxes - find out from the manual @Madis@
if ALGO=0
	caption 		db 'Kmmel Mandelbrot Benchmark V 0.53 MT FPU', 0
end if
if ALGO=1
	caption 		db 'Kmmel Mandelbrot Benchmark V 0.53 MT SSE2', 0
end if
if ALGO=2
	caption 		db 'Kmmel Mandelbrot Benchmark V 0.53 MT SSE3', 0
end if
if ALGO=3
	caption 		db 'Kmmel Mandelbrot Benchmark V 0.53 MT SSE3 CO', 0
end if
_title			db 'KMB Benchmark', 0
_class			db 'FASMWIN32', 0

align 16
frameconfig: FrameArray \
	-1.262898230274756450, -1.261965073162836240, 0.407690116544977221, \	; Lightning Apple
	-0.734132991920317224, -0.734532991920317224, 0.207761617861443873, \	; Freaky Apple
	-0.743640116609113000, -0.743148855074946213, 0.131657512678419047, \	; Circles
	-0.632850590263073443, -0.631963003010647450, 0.400622336442401417, \	; Twirl
	-0.634761297711484929, -0.610764951700536574, 0.384163093347149487, \	; Wing
	-0.740767883072350530, -0.733269024943929093, 0.204531355452692548, \	; Star
	-0.11022029, -0.095409, 0.948442, \										; Branches
	-2.25, 2.25, -2.25														; Standard Mandelbrot Set

MAX_FRAMES = (($ - frameconfig)/24) - 1

Colour_Table_init: ColourArray \
	0,132,212,212,\
	0,120,208,208,\
	0,108,204,204,\
	0,96,200,200,\
	0,84,196,196,\
	0,60,196,188,\
	0,56,192,176,\
	0,52,188,164,\
	0,48,184,152,\
	0,44,180,140,\
	0,40,176,128,\
	0,36,172,116,\
	0,32,168,104,\
	0,28,164,92,\
	0,24,160,80,\
	0,20,156,68,\
	0,16,152,56,\
	0,12,148,44,\
	0,8,144,32,\
	0,4,140,20,\
	0,16,140,20,\
	0,28,136,28,\
	0,40,132,36,\
	0,52,128,44,\
	0,64,124,52,\
	0,76,120,60,\
	0,88,116,68,\
	0,100,112,76,\
	0,112,108,84,\
	0,124,104,92,\
	0,136,100,100,\
	0,148,96,108,\
	0,160,92,116,\
	0,172,88,124,\
	0,184,84,132,\
	0,196,80,140,\
	0,200,92,140,\
	0,204,104,148,\
	0,208,116,156,\
	0,212,128,164,\
	0,216,140,172,\
	0,220,152,180,\
	0,224,164,188,\
	0,228,176,196,\
	0,232,188,204,\
	0,236,200,212,\
	0,240,212,220,\
	0,244,224,228,\
	0,248,236,236,\
	0,252,248,244,\
	0,252,252,252,\
	0,240,248,248,\
	0,228,244,244,\
	0,216,240,240,\
	0,204,236,236,\
	0,192,232,232,\
	0,180,228,228,\
	0,168,224,224,\
	0,156,220,220,\
	0,144,216,216,\
	0,132,212,212,\
	0,120,208,208,\
	0,108,204,204,\
	0,96,200,200,\
	0,84,196,196,\
	0,60,196,188,\
	0,56,192,176,\
	0,52,188,164,\
	0,48,184,152,\
	0,44,180,140,\
	0,40,176,128,\
	0,36,172,116,\
	0,32,168,104,\
	0,28,164,92,\
	0,24,160,80,\
	0,20,156,68,\
	0,16,152,56,\
	0,12,148,44,\
	0,8,144,32,\
	0,4,140,20,\
	0,16,140,20,\
	0,28,136,28,\
	0,40,132,36,\
	0,52,128,44,\
	0,64,124,52,\
	0,76,120,60,\
	0,88,116,68,\
	0,100,112,76,\
	0,112,108,84,\
	0,124,104,92,\
	0,136,100,100,\
	0,148,96,108,\
	0,160,92,116,\
	0,172,88,124,\
	0,184,84,132,\
	0,196,80,140,\
	0,200,92,140,\
	0,204,104,148,\
	0,208,116,156,\
	0,212,128,164,\
	0,216,140,172,\
	0,220,152,180,\
	0,224,164,188,\
	0,228,176,196,\
	0,232,188,204,\
	0,236,200,212,\
	0,240,212,220,\
	0,244,224,228,\
	0,248,236,236,\
	0,252,248,244,\
	0,252,252,252,\
	0,240,248,248,\
	0,228,244,244,\
	0,216,240,240,\
	0,204,236,236,\
	0,192,232,232,\
	0,180,228,228,\
	0,168,224,224,\
	0,156,220,220,\
	0,144,216,216,\
	0,132,212,212,\
	0,120,208,208,\
	0,108,204,204,\
	0,96,200,200,\
	0,84,196,196,\
	0,60,196,188,\
	0,56,192,176,\
	0,52,188,164,\
	0,48,184,152,\
	0,44,180,140,\
	0,0,176,128,\
	0,36,172,116,\
	0,32,168,104,\
	0,28,164,92,\
	0,24,160,80,\
	0,20,156,68,\
	0,16,152,56,\
	0,12,148,44,\
	0,8,144,32,\
	0,4,140,20,\
	0,16,140,20,\
	0,28,136,28,\
	0,40,132,36,\
	0,52,128,44,\
	0,64,124,52,\
	0,76,120,60,\
	0,88,116,68,\
	0,100,112,76,\
	0,112,108,84,\
	0,124,104,92,\
	0,136,100,100,\
	0,148,96,108,\
	0,160,92,116,\
	0,172,88,124,\
	0,184,84,132,\
	0,196,80,140,\
	0,200,92,140,\
	0,204,104,148,\
	0,208,116,156,\
	0,212,128,164,\
	0,216,140,172,\
	0,220,152,180,\
	0,224,164,188,\
	0,228,176,196,\
	0,232,188,204,\
	0,236,200,212,\
	0,240,212,220,\
	0,244,224,228,\
	0,248,236,236,\
	0,252,248,244,\
	0,252,252,252,\
	0,240,248,248,\
	0,228,244,244,\
	0,216,240,240,\
	0,204,236,236,\
	0,192,232,232,\
	0,180,228,228,\
	0,168,224,224,\
	0,156,220,220,\
	0,144,216,216,\
	0,132,212,212,\
	0,120,208,208,\
	0,108,204,204,\
	0,96,200,200,\
	0,84,196,196,\
	0,60,196,188,\
	0,56,192,176,\
	0,52,188,164,\
	0,48,184,152,\
	0,44,180,140,\
	0,40,176,128,\
	0,36,172,116,\
	0,32,168,104,\
	0,28,164,92,\
	0,24,160,80,\
	0,20,156,68,\
	0,16,152,56,\
	0,12,148,44,\
	0,8,144,32,\
	0,4,140,20,\
	0,16,140,20,\
	0,28,136,28,\
	0,40,132,36,\
	0,52,128,44,\
	0,64,124,52,\
	0,76,120,60,\
	0,88,116,68,\
	0,100,112,76,\
	0,112,108,84,\
	0,124,104,92,\
	0,136,100,100,\
	0,148,96,108,\
	0,160,92,116,\
	0,172,88,124,\
	0,184,84,132,\
	0,196,80,140,\
	0,200,92,140,\
	0,204,104,148,\
	0,208,116,156,\
	0,212,128,164,\
	0,216,140,172,\
	0,220,152,180,\
	0,224,164,188,\
	0,228,176,196,\
	0,232,188,204,\
	0,236,200,212,\
	0,240,212,220,\
	0,244,224,228,\
	0,248,236,236,\
	0,252,248,244,\
	0,252,252,252,\
	0,240,248,248,\
	0,228,244,244,\
	0,216,240,240,\
	0,204,236,236,\
	0,192,232,232,\
	0,180,228,228,\
	0,168,224,224,\
	0,156,220,220,\
	0,144,216,216

buffer			rb 10
mainhwnd		dd ?			; the main window handle
flags			dd	?			; seems to be something for the message window...
hinstance		dd ?			; Our programs instance
hdc				dd ?			; handle to device context

threadhandles	dd ?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,? 		; <f0dder> array for WaitForMultipleObjectsEx
tId				dd ?
frame_counter	dd ?
repeat_counter  dd ?
current_line	dd ?			; <f0dder> the (GLOBAL!) current line-position
current_line1	db ?			; <f0dder> the (GLOBAL!) current line-position

Colour_Table	rb (240 * 4 * 10 + 4)	; reserve 2400 palette entries, each 32bit + one for black
